@@ -1,6 +1,7 @@
 from __future__ import division
 
 import copy
+import warnings
 import yaml
 
 import numpy as np
@@ -15,6 +16,40 @@ except ImportError:
     _cv2_available = False
 
 
+def pil_to_cv2_interpolation(interpolation):
+    if isinstance(interpolation, str):
+        interpolation = interpolation.lower()
+        if interpolation == 'nearest':
+            cv_interpolation = cv2.INTER_NEAREST
+        elif interpolation == 'bilinear':
+            cv_interpolation = cv2.INTER_LINEAR
+        elif interpolation == 'bicubic':
+            cv_interpolation = cv2.INTER_CUBIC
+        elif interpolation == 'lanczos':
+            cv_interpolation = cv2.INTER_LANCZOS4
+        else:
+            raise ValueError(
+                'Not valid Interpolation. '
+                'Valid interpolation methods are '
+                'nearest, bilinear, bicubic and lanczos.')
+    else:
+        if interpolation == PIL.Image.NEAREST:
+            cv_interpolation = cv2.INTER_NEAREST
+        elif interpolation == PIL.Image.BILINEAR:
+            cv_interpolation = cv2.INTER_LINEAR
+        elif interpolation == PIL.Image.BICUBIC:
+            cv_interpolation = cv2.INTER_CUBIC
+        elif interpolation == PIL.Image.LANCZOS:
+            cv_interpolation = cv2.INTER_LANCZOS4
+        else:
+            raise ValueError(
+                'Not valid Interpolation. '
+                'Valid interpolation methods are '
+                'PIL.Image.NEAREST, PIL.Image.BILINEAR, '
+                'PIL.Image.BICUBIC and PIL.Image.LANCZOS.')
+    return cv_interpolation
+
+
 def format_mat(x, precision):
     return ("[%s]" % (
         np.array2string(x, precision=precision,
@@ -22,6 +57,29 @@ def format_mat(x, precision):
             .replace("[", "").replace("]", "").replace("\n", "\n        ")))
 
 
+def get_valid_roi(height, width, roi):
+    y1, x1, y2, x2 = roi
+    y1 = max(0, y1)
+    x1 = max(0, x1)
+    y2 = min(height, y2)
+    x2 = min(width, x2)
+    roi_height = y2 - y1
+    roi_width = x2 - x1
+
+    # ROI with non-positive height or width is
+    # considered the same as full resolution.
+    if x1 == 0 and y1 == 0 \
+       and roi_width == 0 and roi_height == 0:
+        return [0, 0, height, width]
+    elif roi_width <= 0 or roi_height <= 0:
+        warnings.warn(
+            "Invalid ROI, [left: {}, top: {}, right: {}, bottom: {}]".format(
+                roi[0], roi[1], roi[2], roi[3]))
+        return [0, 0, height, width]
+    else:
+        return [y1, x1, y2, x2]
+
+
 class PinholeCameraModel(object):
 
     """A Pinhole Camera Model
@@ -44,7 +102,7 @@ class PinholeCameraModel(object):
     D : numpy.ndarray
         distortion.
     roi : None or list[float]
-        [left_y, left_x, right_y, right_x] order.
+        [top, left, bottom, right] order.
     tf_frame : None or str
         tf frame. This is for ROS compatibility.
     stamp : None
@@ -88,7 +146,6 @@ class PinholeCameraModel(object):
         self._height = image_height
         self._full_width = full_width or self._width
         self._full_height = full_height or self._height
-        self._aspect = 1.0 * self.width / self.height
         self.K = K
         self.D = D
         self.R = R
@@ -103,8 +160,6 @@ class PinholeCameraModel(object):
             self._full_P = full_P
         else:
             self._full_P = self.P.copy()
-        self._fovx = 2.0 * np.rad2deg(np.arctan(self.width / (2.0 * self.fx)))
-        self._fovy = 2.0 * np.rad2deg(np.arctan(self.height / (2.0 * self.fy)))
         self._binning_x = binning_x
         self._binning_y = binning_y
         self._roi = roi or [0, 0, self._height, self._width]
@@ -123,6 +178,9 @@ class PinholeCameraModel(object):
         K = self.full_K.copy()
         P = self.full_P.copy()
         # Adjust K and P for binning and ROI
+        if self._target_size is not None:
+            self._binning_x = (x2 - x1) / self._target_size[0]
+            self._binning_y = (y2 - y1) / self._target_size[1]
         K[0, 0] /= self._binning_x
         K[1, 1] /= self._binning_y
         K[0, 2] = (K[0, 2] - x1) / self._binning_x
@@ -135,6 +193,9 @@ class PinholeCameraModel(object):
         self.P = P
         self._width = x2 - x1
         self._height = y2 - y1
+        self._aspect = 1.0 * self.width / self.height
+        self._fovx = 2.0 * np.rad2deg(np.arctan(self.width / (2.0 * self.fx)))
+        self._fovy = 2.0 * np.rad2deg(np.arctan(self.height / (2.0 * self.fy)))
 
     @property
     def width(self):
@@ -505,7 +566,7 @@ class PinholeCameraModel(object):
         Returns
         -------
         self._roi : None or list[float]
-            [left_y, left_x, right_y, right_x] order.
+            [top, left, bottom, right] order.
         """
         return self._roi
 
@@ -516,9 +577,9 @@ class PinholeCameraModel(object):
         Parameters
         ----------
         roi : list[float]
-            [left_y, left_x, right_y, right_x] order.
+            [top, left, bottom, right] order.
         """
-        self._roi = roi
+        self._roi = get_valid_roi(self._full_height, self._full_width, roi)
         self._adjust()
 
     @property
@@ -772,23 +833,40 @@ class PinholeCameraModel(object):
             # opencv format
             image_width = data['image_width']
             image_height = data['image_height']
-            K = data['camera_matrix']['data']
-            P = data['projection_matrix']['data']
-            R = data['rectification_matrix']['data']
-            D = data['distortion_coefficients']['data']
+            K = np.array(
+                data['camera_matrix']['data'],
+                dtype=np.float32).reshape(3, 3)
+            P = np.array(
+                data['projection_matrix']['data'],
+                dtype=np.float32).reshape(3, 4)
+            R = np.array(
+                data['rectification_matrix']['data'],
+                dtype=np.float32).reshape(3, 3)
+            D = np.array(
+                data['distortion_coefficients']['data'],
+                dtype=np.float32)
+
             distortion_model = 'plumb_bob'
             if 'camera_name' in data:
-                name = data['camera_name']
+                name = data['camera_name'] or ''
             else:
                 name = ''
         elif 'width' in data:
             # ROS yaml format
             image_width = data['width']
             image_height = data['height']
-            K = data['K']
-            P = data['P']
-            R = data['R']
-            D = data['D']
+            K = np.array(
+                data['K'],
+                dtype=np.float32).reshape(3, 3)
+            P = np.array(
+                data['P'],
+                dtype=np.float32).reshape(3, 4)
+            R = np.array(
+                data['R'],
+                dtype=np.float32).reshape(3, 3)
+            D = np.array(
+                data['D'],
+                dtype=np.float32)
             distortion_model = data['distortion_model']
             name = ''
         else:
@@ -799,28 +877,36 @@ class PinholeCameraModel(object):
         if 'binning_y' in data:
             binning_y = 1 if data['binning_y'] == 0 else data['binning_y']
 
-        # ROI all zeros is considered the same as full resolution
+        roi_width = image_width
+        roi_height = image_height
         if 'roi' in data:
             x_offset = data['roi']['x_offset']
             y_offset = data['roi']['y_offset']
             roi_width = data['roi']['width']
             roi_height = data['roi']['height']
-            if x_offset == 0 \
-               and y_offset == 0 \
-               and roi_width == 0 \
-               and roi_height == 0:
-                roi_width = image_width
-                roi_height = image_height
-            roi = [y_offset,
-                   x_offset,
-                   y_offset + roi_height,
-                   x_offset + roi_width]
+            roi = get_valid_roi(
+                image_height, image_width,
+                [y_offset,
+                 x_offset,
+                 y_offset + roi_height,
+                 x_offset + roi_width])
+            roi_width = roi[3] - roi[1]
+            roi_height = roi[2] - roi[0]
 
+        full_K = K.copy()
+        full_P = P.copy()
         return PinholeCameraModel(
-            image_height, image_width,
-            K, P, R, D, roi=roi,
+            roi_height, roi_width,
+            K, P, R, D,
+            roi=roi,
             distortion_model=distortion_model,
-            name=name, binning_x=binning_x, binning_y=binning_y)
+            name=name,
+            full_K=full_K,
+            full_P=full_P,
+            full_width=image_width,
+            full_height=image_height,
+            binning_x=binning_x,
+            binning_y=binning_y)
 
     @staticmethod
     def from_camera_info(camera_info_msg):
@@ -863,23 +949,22 @@ class PinholeCameraModel(object):
             binning_y = camera_info_msg.binning_y
 
         raw_roi = copy.copy(camera_info_msg.roi)
-        # ROI all zeros is considered the same as full resolution
-        if (raw_roi.x_offset == 0 and raw_roi.y_offset == 0 and
-                raw_roi.width == 0 and raw_roi.height == 0):
-            raw_roi.width = image_width
-            raw_roi.height = image_height
+        roi = get_valid_roi(
+            image_height, image_width,
+            [raw_roi.y_offset,
+             raw_roi.x_offset,
+             raw_roi.y_offset + raw_roi.height,
+             raw_roi.x_offset + raw_roi.width])
+        roi_width = roi[3] - roi[1]
+        roi_height = roi[2] - roi[0]
 
-        roi = [raw_roi.y_offset,
-               raw_roi.x_offset,
-               raw_roi.y_offset + raw_roi.height,
-               raw_roi.x_offset + raw_roi.width]
         tf_frame = camera_info_msg.header.frame_id
         stamp = camera_info_msg.header.stamp
 
         full_K = K.copy()
         full_P = P.copy()
         return PinholeCameraModel(
-            raw_roi.height, raw_roi.width,
+            roi_height, roi_width,
             K, P, R, D,
             roi,
             tf_frame,
@@ -892,6 +977,38 @@ class PinholeCameraModel(object):
             binning_x=binning_x,
             binning_y=binning_y)
 
+    def rectify_image(self, raw_img,
+                      interpolation=PIL.Image.BILINEAR):
+        """Rectify input raw image.
+
+        Parameters
+        ----------
+        raw_img : numpy.ndarray
+            raw image.
+        interpolation : int
+            interpolation method.
+            You can specify, PIL.Image.NEAREST, PIL.Image.BILINEAR,
+            PIL.Image.BICUBIC and PIL.Image.LANCZOS.
+
+        Returns
+        -------
+        rectified_img : numpy.ndarray
+            rectified image.
+        """
+        if _cv2_available is False:
+            raise RuntimeError('CV2 are not enabled. Currently '
+                               'only support cv2 rectification.')
+        mapx = np.ndarray(shape=(self.height, self.width, 1),
+                          dtype='float32')
+        mapy = np.ndarray(shape=(self.height, self.width, 1),
+                          dtype='float32')
+        cv2.initUndistortRectifyMap(
+            self.K, self.D, self.R, self.P,
+            (self.width, self.height),
+            cv2.CV_32FC1, mapx, mapy)
+        cv_interpolation = pil_to_cv2_interpolation(interpolation)
+        return cv2.remap(raw_img, mapx, mapy, cv_interpolation)
+
     def crop_resize_camera_info(self, target_size, roi=None):
         """Return cropped and resized region's camera model
 
@@ -920,6 +1037,8 @@ class PinholeCameraModel(object):
                 "Invalid ROI, [left: {}, top: {}, right: {}, bottom: {}]".
                 format(roi[0], roi[1], roi[2], roi[3]))
 
+        target_size = self._calc_resize_image_size_keeping_aspect_ratio(
+            target_size)
         binning_x = roi_width / target_size[0]
         binning_y = roi_height / target_size[1]
 
@@ -1011,14 +1130,7 @@ class PinholeCameraModel(object):
         cropped_img = img[y1:y2, x1:x2]
         out = np.empty(out_shape, dtype=img.dtype)
         if use_cv2 and _cv2_available:
-            if interpolation == PIL.Image.NEAREST:
-                cv_interpolation = cv2.INTER_NEAREST
-            elif interpolation == PIL.Image.BILINEAR:
-                cv_interpolation = cv2.INTER_LINEAR
-            elif interpolation == PIL.Image.BICUBIC:
-                cv_interpolation = cv2.INTER_CUBIC
-            elif interpolation == PIL.Image.LANCZOS:
-                cv_interpolation = cv2.INTER_LANCZOS4
+            cv_interpolation = pil_to_cv2_interpolation(interpolation)
             out[:] = cv2.resize(cropped_img, self._target_size,
                                 interpolation=cv_interpolation)
         else:
@@ -1445,3 +1557,32 @@ class PinholeCameraModel(object):
                 result.append(False)
 
         return result
+
+    def _calc_resize_image_size_keeping_aspect_ratio(
+            self, target_size):
+        """Calculate resized image size keeping aspect ratio.
+
+        Parameters
+        ----------
+        target_size : tuple(int or None)
+            values of (width, height).
+            If width or height is None,
+            calculate it from aspect ratio.
+
+        Returns
+        -------
+        resized_size : tuple(int)
+            resized target size.
+        """
+        width, height = target_size
+        if width is not None and height is not None:
+            return (width, height)
+        if width is None and height is None:
+            raise ValueError('Only width or height should be specified.')
+        if width:
+            height = width * self._full_height / self._full_width
+        else:
+            width = height * self._full_width / self._full_height
+        height = int(height)
+        width = int(width)
+        return (width, height)
