@@ -23,6 +23,11 @@ class TestPinholeCameraModel(unittest.TestCase):
     def setUpClass(cls):
         cls.cm = PinholeCameraModel.from_fovy(45, 480, 640)
 
+    def test_rectify_image(self):
+        cm = PinholeCameraModel.from_yaml_file(kinect_v2_camera_info())
+        img = kinect_v2_image()
+        cm.rectify_image(img)
+
     def test_crop_resize_camra_info(self):
         cropped_resized_cm = self.cm.crop_resize_camera_info(
             target_size=[100, 200], roi=[0, 0, 100, 150])
@@ -192,9 +197,19 @@ class TestPinholeCameraModel(unittest.TestCase):
 
     def test__target_size(self):
         cm = PinholeCameraModel.from_yaml_file(kinect_v2_camera_info())
+        org_roi = copy.deepcopy(cm.roi)
         cm.target_size = (640, 480)
         self.assertEqual(cm.target_size, (640, 480))
 
+        resize_K = cm.K.copy()
+        resize_P = cm.P.copy()
+
+        cm.roi = [0, 0, 100, 100]
+        cm.target_size = (640, 480)
+        cm.roi = org_roi
+        testing.assert_equal(resize_K, cm.K)
+        testing.assert_equal(resize_P, cm.P)
+
     def test__binning_x(self):
         cm = PinholeCameraModel.from_yaml_file(kinect_v2_camera_info())
         cm.roi = [0, 0, 100, 100]
